#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-

"""
web server.
__author__ = "Valentin Kuznetsov"
"""
from __future__ import print_function

from builtins import str as newstr, bytes, map
from future.utils import viewitems, viewvalues

# system modules
import collections.abc
import json
import os
import pprint
import sys
import time
from copy import deepcopy
from datetime import datetime

# cherrypy modules
import cherrypy
from cherrypy import config as cherryconf
from cherrypy import expose, response, tools
from cherrypy.lib.static import serve_file

# import WMCore itself to determine path of modules
import WMCore
from Utils.CertTools import getKeyCertFromEnv
from WMCore.REST.Auth import get_user_info
# WMCore modules
from WMCore.ReqMgr.DataStructs.RequestStatus import ACTIVE_STATUS
from WMCore.ReqMgr.DataStructs.RequestStatus import REQUEST_STATE_TRANSITION, REQUEST_HUMAN_STATES
from WMCore.ReqMgr.DataStructs.RequestStatus import get_modifiable_properties, get_protected_properties
from WMCore.ReqMgr.Tools.cms import lfn_bases, lfn_unmerged_bases
from WMCore.ReqMgr.Tools.cms import releases, dashboardActivities
from WMCore.ReqMgr.Tools.cms import site_white_list, site_black_list
from WMCore.ReqMgr.Tools.cms import web_ui_names, SITE_CACHE, PNN_CACHE
from WMCore.ReqMgr.Utils.Validation import get_request_template_from_type
# ReqMgrSrv modules
from WMCore.ReqMgr.Web.tools import exposecss, exposejs, TemplatedPage
from WMCore.ReqMgr.Web.utils import gen_color
from WMCore.ReqMgr.Web.utils import json2table, json2form, genid, checkargs, tstamp, sort, reorder_list
from WMCore.Services.LogDB.LogDB import LogDB
# new reqmgr2 APIs
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.Services.pycurl_manager import RequestHandler
from WMCore.WMSpec.StdSpecs.StdBase import StdBase
from WMCore.Cache.GenericDataCache import MemoryCacheStruct


def getdata(url, params, headers=None):
    "Helper function to get data from the service"
    ckey, cert = getKeyCertFromEnv()
    mgr = RequestHandler()
    res = mgr.getdata(url, params=params, headers=headers, ckey=ckey, cert=cert)
    return json.loads(res)


def sort_bold(docs):
    "Return sorted list of bold items from provided doc list"
    return ', '.join(['<b>%s</b>' % m for m in sorted(docs)])


def set_headers(itype, size=0):
    """
    Set response header Content-type (itype) and Content-Length (size).
    """
    if size > 0:
        response.headers['Content-Length'] = size
    response.headers['Content-Type'] = itype
    response.headers['Expires'] = 'Sat, 14 Oct 2027 00:59:30 GMT'


def set_no_cache_flags():
    "Set cherrypy flags to prevent caching"
    cherrypy.response.headers['Cache-Control'] = 'no-cache'
    cherrypy.response.headers['Pragma'] = 'no-cache'
    cherrypy.response.headers['Expires'] = 'Sat, 01 Dec 2001 00:00:00 GMT'


def set_cache_flags():
    "Set cherrypy flags to prevent caching"
    headers = cherrypy.response.headers
    for key in ['Cache-Control', 'Pragma']:
        if key in headers:
            del headers[key]


def minify(content):
    """
    Remove whitespace in provided content.
    """
    content = content.replace('\n', ' ')
    content = content.replace('\t', ' ')
    content = content.replace('   ', ' ')
    content = content.replace('  ', ' ')
    return content


def menus():
    "Return dict of menus"
    items = ['home', 'create', 'approve', 'assign', 'batches']
    return items


def request_attr(doc, attrs=None):
    "Return request attributes/values in separate document"
    if not attrs:
        attrs = ['RequestName', 'Requestdate', 'Inputdataset', \
                 'Prepid', 'Group', 'Requestor', 'RequestDate', \
                 'RequestStatus']
    rdict = {}
    for key in attrs:
        if key in doc:
            if key == 'RequestDate':
                tval = doc[key]
                if isinstance(tval, list):
                    while len(tval) < 9:
                        tval.append(0)
                    # we do not know if dayling savings time was in effect or not
                    tval[-1] = -1
                    gmt = time.gmtime(time.mktime(tuple(tval)))
                    rdict[key] = time.strftime("%Y-%m-%d %H:%M:%S GMT", gmt)
                else:
                    rdict[key] = tval
            else:
                rdict[key] = doc[key]
    return rdict


def spec_list(root):
    "Return list of specs from given root directory"
    specs = []
    for fname in os.listdir(root):
        if not fname.endswith('.py') or fname == '__init__.py':
            continue
        sname = fname.split('.')[0]
        clsName = "%sWorkloadFactory" % sname
        with open(os.path.join(root, fname)) as fd:
            if clsName in fd.read():
                specs.append(sname)
    return specs


def user():
    """
    Return user name associated with this instance.
    """
    try:
        return get_user_info()['login']
    except:
        return 'testuser'


def user_dn():
    "Return user DN"
    try:
        return get_user_info()['dn']
    except:
        return '/CN/bla/foo'


def check_scripts(scripts, resource, path):
    """
    Check a script is known to the resource map
    and that the script actually exists
    """
    for script in scripts:
        if script not in resource:
            spath = os.path.normpath(os.path.join(path, script))
            if os.path.isfile(spath):
                resource.update({script: spath})
    return scripts


def _map_configcache_url(tConfigs, baseURL, configIDName, configID, taskName=""):
    if configIDName.endswith('ConfigCacheID') and configID is not None:
        url = "%s/reqmgr_config_cache/%s/configFile" % (baseURL, configID)
        prefix = "%s: " % taskName if taskName else ""
        task = "%s%s: %s " % (prefix, configIDName, configID)
        tConfigs.setdefault(task, url)
    return


def tasks_configs(docs, html=False):
    "Helper function to provide mapping between tasks and configs"
    if not isinstance(docs, list):
        docs = [docs]
    tConfigs = {}
    for doc in docs:
        name = doc.get('RequestName', '')
        if "TaskChain" in doc:
            chainTypeFlag = True
            ctype = "Task"
        elif "StepChain" in doc:
            chainTypeFlag = True
            ctype = "Step"
        else:
            chainTypeFlag = False
            ctype = None

        curl = doc.get('ConfigCacheUrl', 'https://cmsweb.cern.ch/couchdb')
        if curl == None or curl == "none":
            curl = 'https://cmsweb.cern.ch/couchdb'
        if not name:
            continue
        for key, val in viewitems(doc):
            _map_configcache_url(tConfigs, curl, key, val)
            if chainTypeFlag and key.startswith(ctype) and isinstance(val, dict):
                for kkk in val:
                    # append task/step number and name
                    keyStr = "%s: %s" % (key, val.get("%sName" % ctype, ''))
                    _map_configcache_url(tConfigs, curl, kkk, val[kkk], keyStr)
    if html:
        out = '<fieldset><legend>Config Cache List</legend><ul>'
        for task in sorted(tConfigs):
            out += '<li><a href="%s" target="config_page">%s</a></li>' % (tConfigs[task], task)
        out += '</ul></fieldset>'
        return out
    return tConfigs


def state_transition(docs):
    "Helper function to provide mapping between tasks and configs"
    if not isinstance(docs, list):
        docs = [docs]

    out = '<fieldset><legend>State Transition</legend><ul>'
    multiDocFlag = True if len(docs) > 1 else False
    for doc in docs:
        name = doc.get('RequestName', '')
        sTransition = doc.get('RequestTransition', '')

        if not name:
            continue
        if multiDocFlag:
            out += '%s<br />' % name
        for sInfo in sTransition:
            out += '<li><b>%s</b>: %s UTC <b>DN</b>: %s</li>' % (sInfo["Status"],
                                                                 datetime.utcfromtimestamp(
                                                                     sInfo["UpdateTime"]).strftime('%Y-%m-%d %H:%M:%S'),
                                                                 sInfo["DN"])
    out += '</ul></fieldset>'
    return out


def priority_transition(docs):
    "create html for priority transition format"
    if not isinstance(docs, list):
        docs = [docs]

    out = '<fieldset><legend>Priority Transition</legend><ul>'
    multiDocFlag = True if len(docs) > 1 else False
    for doc in docs:
        name = doc.get('RequestName', '')
        pTransition = doc.get('PriorityTransition', '')

        if not name:
            continue
        if multiDocFlag:
            out += '%s<br />' % name
        for pInfo in pTransition:
            out += '<li><b>%s</b>: %s UTC <b>DN</b>: %s</li>' % (pInfo["Priority"],
                                                                 datetime.utcfromtimestamp(
                                                                     pInfo["UpdateTime"]).strftime('%Y-%m-%d %H:%M:%S'),
                                                                 pInfo["DN"])
    out += '</ul></fieldset>'
    return out


# code taken from
# http://stackoverflow.com/questions/1254454/fastest-way-to-convert-a-dicts-keys-values-from-unicode-to-str
def toString(data):
    if isinstance(data, (newstr, bytes)):
        return str(data)
    elif isinstance(data, collections.abc.Mapping):
        return dict(list(map(toString, viewitems(data))))
    elif isinstance(data, collections.abc.Iterable):
        return type(data)(list(map(toString, data)))
    else:
        return data


def getPropValueMap():
    """
    Return all possible values for some assignment arguments
    """
    prop_value_map = {'CMSSWVersion': releases(),
                      'SiteWhitelist': SITE_CACHE.getData(),
                      'SiteBlacklist': SITE_CACHE.getData(),
                      'SubscriptionPriority': ['Low', 'Normal', 'High'],
                      'CustodialSites': PNN_CACHE.getData(),
                      'NonCustodialSites': PNN_CACHE.getData(),
                      'MergedLFNBase': lfn_bases(),
                      'UnmergedLFNBase': lfn_unmerged_bases(),
                      'TrustPUSitelists': [True, False],
                      'TrustSitelists': [True, False],
                      'Dashboard': dashboardActivities()}
    return prop_value_map


class ReqMgrService(TemplatedPage):
    """
    Request Manager web service class
    """

    def __init__(self, app, config, mount):
        self.base = config.base
        self.rootdir = '/'.join(WMCore.__file__.split('/')[:-1])
        if config and not isinstance(config, dict):
            web_config = config.dictionary_()
        if not config:
            web_config = {'base': self.base}
        TemplatedPage.__init__(self, web_config)
        imgdir = os.environ.get('RM_IMAGESPATH', os.getcwd() + '/images')
        self.imgdir = web_config.get('imgdir', imgdir)
        cssdir = os.environ.get('RM_CSSPATH', os.getcwd() + '/css')
        self.cssdir = web_config.get('cssdir', cssdir)
        jsdir = os.environ.get('RM_JSPATH', os.getcwd() + '/js')
        self.jsdir = web_config.get('jsdir', jsdir)
        spdir = os.environ.get('RM_SPECPATH', os.getcwd() + '/specs')
        self.spdir = web_config.get('spdir', spdir)
        # read scripts area and initialize data-ops scripts
        self.sdir = os.environ.get('RM_SCRIPTS', os.getcwd() + '/scripts')
        self.sdir = web_config.get('sdir', self.sdir)
        self.sdict_thr = web_config.get('sdict_thr', 600)  # put reasonable 10 min interval
        self.sdict = {'ts': time.time()}  # placeholder for data-ops scripts
        self.update_scripts(force=True)

        # To be filled at run time
        self.cssmap = {}
        self.jsmap = {}
        self.imgmap = {}

        std_specs_dir = os.path.join(self.rootdir, 'WMSpec/StdSpecs')
        self.std_specs = spec_list(std_specs_dir)
        self.std_specs.sort()

        # Update CherryPy configuration
        mime_types = ['text/css']
        mime_types += ['application/javascript', 'text/javascript',
                       'application/x-javascript', 'text/x-javascript']
        cherryconf.update({'tools.encode.on': True,
                           'tools.gzip.on': True,
                           'tools.gzip.mime_types': mime_types,
                           })
        self._cache = {}

        # initialize access to reqmgr2 APIs
        self.reqmgr_url = config.reqmgr.reqmgr2_url
        self.reqmgr = ReqMgr(self.reqmgr_url)
        # only gets current view (This might cause to reponse time much longer,
        # If upto date view is not needed overwrite Fale)
        self.reqmgr._noStale = True

        # get fields which we'll use in templates
        cdict = config.reqmgr.dictionary_()
        self.couch_url = cdict.get('couch_host', '')
        self.couch_dbname = cdict.get('couch_reqmgr_db', '')
        self.couch_wdbname = cdict.get('couch_workload_summary_db', '')
        self.acdc_url = cdict.get('acdc_host', '')
        self.acdc_dbname = cdict.get('acdc_db', '')
        self.configcache_url = cdict.get('couch_config_cache_url', self.couch_url)
        self.dbs_url = cdict.get('dbs_url', '')
        self.dqm_url = cdict.get('dqm_url', '')
        self.sw_ver = cdict.get('default_sw_version', 'CMSSW_7_6_1')
        self.sw_arch = cdict.get('default_sw_scramarch', 'slc6_amd64_gcc493')

        # LogDB holder
        centralurl = cdict.get("central_logdb_url", "")
        identifier = cdict.get("log_reporter", "reqmgr2")
        self.logdb = LogDB(centralurl, identifier)

        # local team cache which will request data from wmstats
        base, uri = self.reqmgr_url.split('://')
        base_url = '%s://%s' % (base, uri.split('/')[0])
        self.wmstatsurl = cdict.get('wmstats_url', '%s/wmstatsserver' % base_url)
        if not self.wmstatsurl:
            raise Exception('ReqMgr2 configuration file does not provide wmstats url')
        # cache team information for 2 hours to limit wmstatsserver API calls
        self.TEAM_CACHE = MemoryCacheStruct(7200, self.refreshTeams)

        # fetch assignment arguments specification from StdBase
        self.assignArgs = StdBase().getWorkloadAssignArgs()
        self.assignArgs = {key: val['default'] for key, val in viewitems(self.assignArgs)}

    def getTeams(self):
        return self.TEAM_CACHE.getData()

    def refreshTeams(self):
        "Helper function to cache team info from wmstats"
        url = '%s/data/teams' % self.wmstatsurl
        params = {}
        headers = {'Accept': 'application/json'}
        try:
            data = getdata(url, params, headers)
            if 'error' in data:
                print("WARNING: fail to get teams from %s" % url)
                print(data)
            teams = data.get('result', [])
            return teams
        except Exception as exp:
            print("WARNING: fail to get teams from %s" % url)
            print(str(exp))

    def update_scripts(self, force=False):
        "Update scripts dict"
        if force or abs(time.time() - self.sdict['ts']) > self.sdict_thr:
            if os.path.isdir(self.sdir):
                for item in os.listdir(self.sdir):
                    with open(os.path.join(self.sdir, item), 'r') as istream:
                        self.sdict[item.split('.')[0]] = istream.read()
            self.sdict['ts'] = time.time()

    def abs_page(self, tmpl, content):
        """generate abstract page"""
        menu = self.templatepage('menu', menus=menus(), tmpl=tmpl)
        body = self.templatepage('generic', menu=menu, content=content)
        page = self.templatepage('main', content=body, user=user())
        return page

    def page(self, content):
        """
        Provide page wrapped with top/bottom templates.
        """
        return self.templatepage('main', content=content)

    def error(self, content):
        "Generate common error page"
        content = self.templatepage('error', content=content)
        return self.abs_page('error', content)

    @expose
    def index(self):
        """Main page"""
        content = self.templatepage('index', requests=ACTIVE_STATUS, rdict=REQUEST_STATE_TRANSITION)
        return self.abs_page('main', content)

    @expose
    def home(self, **kwds):
        """Main page"""
        return self.index(**kwds)

    ### Request actions ###

    @expose
    @checkargs(['status', 'sort'])
    def assign(self, **kwds):
        """assign page"""
        if not kwds:
            kwds = {}
        if 'status' not in kwds:
            kwds.update({'status': 'assignment-approved'})
        docs = []
        attrs = ['RequestName', 'RequestDate', 'Group', 'Requestor', 'RequestStatus']
        dataResult = self.reqmgr.getRequestByStatus(statusList=[kwds['status']])
        for data in dataResult:
            for val in viewvalues(data):
                docs.append(request_attr(val, attrs))
        sortby = kwds.get('sort', 'status')
        docs = [r for r in sort(docs, sortby)]
        assignDict = deepcopy(self.assignArgs)
        assignDict.update(getPropValueMap())
        assignDict['Team'] = self.getTeams()
        filter_sort = self.templatepage('filter_sort')
        content = self.templatepage('assign', sort=sortby,
                                    filter_sort_table=filter_sort,
                                    sites=SITE_CACHE.getData(),
                                    site_white_list=site_white_list(),
                                    site_black_list=site_black_list(),
                                    user=user(), user_dn=user_dn(), requests=toString(docs),
                                    misc_table=json2table(assignDict, web_ui_names(), "all_attributes"),
                                    misc_json=json2form(assignDict, indent=2, keep_first_value=True))
        return self.abs_page('assign', content)

    @expose
    @checkargs(['status', 'sort'])
    def approve(self, **kwds):
        """
        Approve page: get list of request associated with user DN.
        Fetch their status list from ReqMgr and display if requests
        were seen by data-ops.
        """
        if not kwds:
            kwds = {}
        if 'status' not in kwds:
            kwds.update({'status': 'new'})
        kwds.update({'_nostale': True})
        docs = []
        attrs = ['RequestName', 'RequestDate', 'Group', 'Requestor', 'RequestStatus', 'Campaign']
        dataResult = self.reqmgr.getRequestByStatus(statusList=[kwds['status']])
        for data in dataResult:
            for val in viewvalues(data):
                docs.append(request_attr(val, attrs))
        sortby = kwds.get('sort', 'status')
        docs = [r for r in sort(docs, sortby)]
        filter_sort = self.templatepage('filter_sort')
        content = self.templatepage('approve', requests=toString(docs), date=tstamp(),
                                    sort=sortby, filter_sort_table=filter_sort,
                                    gen_color=gen_color)
        return self.abs_page('approve', content)

    @expose
    def create(self, **kwds):
        """create page"""
        # get list of standard specs from WMCore and new ones from local area
        # loc_specs_dir = os.path.join(self.spdir, 'Specs') # local specs
        # loc_specs = spec_list(loc_specs_dir, 'Specs')
        # all_specs = list(set(self.std_specs + loc_specs))
        # all_specs.sort()
        all_specs = list(self.std_specs)
        spec = kwds.get('form', '')
        if not spec:
            spec = self.std_specs[0]
        # make spec first in all_specs list
        if spec in all_specs:
            all_specs.remove(spec)
        all_specs = [spec] + all_specs
        jsondata = get_request_template_from_type(spec)
        # create templatized page out of provided forms
        self.update_scripts()
        content = self.templatepage('create', table=json2table(jsondata, web_ui_names(), jsondata),
                                    jsondata=json2form(jsondata, indent=2, keep_first_value=True), name=spec,
                                    scripts=[s for s in self.sdict if s != 'ts'],
                                    specs=all_specs)
        return self.abs_page('create', content)

    @expose
    def config(self, name):
        "Fetch config for given request name"
        result = self.reqmgr.getConfig(name)
        if len(result) == 1:
            result = result[0]
        else:
            result = 'Configuration not found for: %s' % name
        return result.replace('\n', '<br/>')

    @expose
    def fetch(self, rid):
        "Fetch document for given id"
        rid = rid.replace('request-', '')
        doc = self.reqmgr.getRequestByNames(rid)
        transitions = []
        tst = time.time()
        # get request tasks
        tasks = self.reqmgr.getRequestTasks(rid)
        if len(doc) == 1:
            try:
                doc = doc[0][rid]
            except:
                pass
            name = doc.get('RequestName', 'NA')
            title = 'Request %s' % name
            status = doc.get('RequestStatus', '')
            transitions = REQUEST_STATE_TRANSITION.get(status, [])
            if status in transitions:
                transitions.remove(status)
            visible_attrs = get_modifiable_properties(status)
            filterout_attrs = get_protected_properties()
            # extend filterout list with "RequestStatus" since it is passed separately
            filterout_attrs.append("RequestStatus")

            for key, val in viewitems(self.assignArgs):
                if not doc.get(key):
                    doc[key] = val

            if visible_attrs == "all_attributes":
                filteredDoc = doc
                for prop in filterout_attrs:
                    if prop in filteredDoc:
                        del filteredDoc[prop]
            else:
                filteredDoc = {}
                for prop in visible_attrs:
                    filteredDoc[prop] = doc.get(prop, "")

            propValueMap = getPropValueMap()
            propValueMap['Team'] = self.getTeams()

            selected = {}
            for prop in propValueMap:
                if prop in filteredDoc:
                    filteredDoc[prop], selected[prop] = reorder_list(propValueMap[prop], filteredDoc[prop])

            content = self.templatepage('doc', title=title, status=status, name=name, rid=rid,
                                        tasks=json2form(tasks, indent=2, keep_first_value=False),
                                        table=json2table(filteredDoc, web_ui_names(), visible_attrs, selected),
                                        jsondata=json2form(doc, indent=2, keep_first_value=False),
                                        doc=json.dumps(doc), time=time,
                                        tasksConfigs=tasks_configs(doc, html=True),
                                        sTransition=state_transition(doc),
                                        pTransition=priority_transition(doc),
                                        transitions=transitions, humanStates=REQUEST_HUMAN_STATES,
                                        ts=tst, user=user(), userdn=user_dn())
        elif len(doc) > 1:
            jsondata = [pprint.pformat(d) for d in doc]
            content = self.templatepage('doc', title='Series of docs: %s' % rid,
                                        table="", jsondata=jsondata, time=time,
                                        tasksConfigs=tasks_configs(doc, html=True),
                                        sTransition=state_transition(doc),
                                        pTransition=priority_transition(doc),
                                        transitions=transitions, humanStates=REQUEST_HUMAN_STATES,
                                        ts=tst, user=user(), userdn=user_dn())
        else:
            doc = 'No request found for name=%s' % rid
        return self.abs_page('request', content)

    @expose
    def record2logdb(self, **kwds):
        """LogDB submission page"""
        print(kwds)
        request = kwds['request']
        msg = kwds['message']
        self.logdb.post(request, msg)
        msg = '<h6>Confirmation</h6>Your request has been entered to LogDB.'
        return self.abs_page('generic', msg)

    @expose
    def requests(self, **kwds):
        """Page showing requests"""
        if not kwds:
            kwds = {}
        if 'status' not in kwds:
            kwds.update({'status': 'acquired'})
        dataResult = self.reqmgr.getRequestByStatus(kwds['status'])
        attrs = ['RequestName', 'RequestDate', 'Group', 'Requestor', 'RequestStatus', 'Campaign']
        docs = []
        for data in dataResult:
            for doc in viewvalues(data):
                docs.append(request_attr(doc, attrs))
        sortby = kwds.get('sort', 'status')
        docs = [r for r in sort(docs, sortby)]
        filter_sort = self.templatepage('filter_sort')
        content = self.templatepage('requests', requests=toString(docs), sort=sortby,
                                    status=kwds['status'], filter_sort_table=filter_sort)
        return self.abs_page('requests', content)

    @expose
    def request(self, **kwargs):
        "Get data example and expose it as json"
        dataset = kwargs.get('uinput', '')
        if not dataset:
            return {'error': 'no input dataset'}
        url = 'https://cmsweb.cern.ch/reqmgr2/data/request?outputdataset=%s' % dataset
        params = {}
        headers = {'Accept': 'application/json'}
        wdata = getdata(url, params, headers)
        wdict = dict(date=time.ctime(), team='Team-A', status='Running', ID=genid(wdata))
        winfo = self.templatepage('workflow', wdict=wdict,
                                  dataset=dataset, code=pprint.pformat(wdata))
        content = self.templatepage('search', content=winfo)
        return self.abs_page('request', content)

    @expose
    def batch(self, **kwds):
        """batch page"""
        # TODO: we need a template for batch attributes
        #       and read it from separate area, like DASMaps
        name = kwds.get('name', '')
        batch = {}
        if name:
            #            batch = self.reqmgr.getBatchesByName(name)
            batch = {'Name': 'Batch1', 'Description': 'Bla-bla', 'Creator': 'valya', 'Group': 'test',
                     'Workflows': ['workflow1', 'workflow2'],
                     'Attributes': {'HeavyIon': ['true', 'false']}}
        attributes = batch.get('Attributes', {})
        workflows = batch.get('Workflows', [])
        description = batch.get('Description', '')
        creator = batch.get('Creator', user_dn())
        content = self.templatepage('batch', name=name,
                                    attributes=json2table(attributes, web_ui_names()),
                                    workflows=workflows, creator=creator,
                                    description=description)
        return self.abs_page('batch', content)

    @expose
    def batches(self, **kwds):
        """Page showing batches"""
        if not kwds:
            kwds = {}
        if 'name' not in kwds:
            kwds.update({'name': ''})
        sortby = kwds.get('sort', 'name')
        #        results = self.reqmgr.getBatchesByName(kwds['name'])
        results = [
            {'Name': 'Batch1', 'Description': 'Bla-bla', 'Creator': 'valya', 'Group': 'test',
             'Workflows': ['workflow1', 'workflow2'],
             'Date': 'Fri Feb 13 10:36:41 EST 2015',
             'Attributes': {'HeavyIon': ['true', 'false']}},
            {'Name': 'Batch2', 'Description': 'lksdjflksjdf', 'Creator': 'valya', 'Group': 'test',
             'Workflows': ['workflow1', 'workflow2'],
             'Date': 'Fri Feb 10 10:36:41 EST 2015',
             'Attributes': {'HeavyIon': ['true', 'false']}},
        ]
        docs = [r for r in sort(results, sortby)]
        filter_sort = self.templatepage('filter_sort')
        content = self.templatepage('batches', batches=docs, sort=sortby,
                                    filter_sort_table=filter_sort)
        return self.abs_page('batches', content)

    ### Aux methods ###

    @expose
    def images(self, *args):
        """
        Serve static images.
        """
        args = list(args)
        check_scripts(args, self.imgmap, self.imgdir)
        mime_types = ['*/*', 'image/gif', 'image/png',
                      'image/jpg', 'image/jpeg']
        accepts = cherrypy.request.headers.elements('Accept')
        for accept in accepts:
            if accept.value in mime_types and len(args) == 1 \
                    and args[0] in self.imgmap:
                image = self.imgmap[args[0]]
                # use image extension to pass correct content type
                ctype = 'image/%s' % image.split('.')[-1]
                cherrypy.response.headers['Content-type'] = ctype
                return serve_file(image, content_type=ctype)

    def serve(self, kwds, imap, idir, datatype='', minimize=False):
        "Serve files for high level APIs (yui/css/js)"
        args = []
        for key, val in viewitems(kwds):
            if key == 'f':  # we only look-up files from given kwds dict
                if isinstance(val, list):
                    args += val
                else:
                    args.append(val)
        scripts = check_scripts(args, imap, idir)
        return self.serve_files(args, scripts, imap, datatype, minimize)

    @exposecss
    @tools.gzip()
    def css(self, **kwargs):
        """
        Serve provided CSS files. They can be passed as
        f=file1.css&f=file2.css
        """
        resource = kwargs.get('resource', 'css')
        if resource == 'css':
            return self.serve(kwargs, self.cssmap, self.cssdir, 'css', True)

    @exposejs
    @tools.gzip()
    def js(self, **kwargs):
        """
        Serve provided JS scripts. They can be passed as
        f=file1.js&f=file2.js with optional resource parameter
        to speficy type of JS files, e.g. resource=yui.
        """
        resource = kwargs.get('resource', 'js')
        if resource == 'js':
            return self.serve(kwargs, self.jsmap, self.jsdir)

    def serve_files(self, args, scripts, resource, datatype='', minimize=False):
        """
        Return asked set of files for JS, YUI, CSS.
        """
        idx = "-".join(scripts)
        if idx not in self._cache:
            data = ''
            if datatype == 'css':
                data = '@CHARSET "UTF-8";'
            for script in args:
                path = os.path.join(sys.path[0], resource[script])
                path = os.path.normpath(path)
                with open(path) as ifile:
                    data = "\n".join([data, ifile.read().replace('@CHARSET "UTF-8";', '')])
            if datatype == 'css':
                set_headers("text/css")
            if minimize:
                self._cache[idx] = minify(data)
            else:
                self._cache[idx] = data
        return self._cache[idx]



#!/usr/bin/env python
"""
_MySQLCore_t

Unit tests for the MySQL version of DBCore.
"""




import unittest
import logging

from WMCore.Database.MySQLCore import MySQLInterface
from WMQuality.TestInit import TestInit

class DBCoreTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Setup logging.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        return

    def tearDown(self):
        """
        _tearDown_

        Nothing to do.
        """
        return

    def testBindSubstitutionA(self):
        """
        _testBindSubstitutionA_

        Verify that bind substition works correctly with similarly named bind
        variables.
        """
        sql = "INSERT INTO RELEASE_VERSIONS (RELEASE_VERSION_ID, RELEASE_VERSION) VALUES (:release_version_id, :release_version)"
        binds = [{"release_version_id": 1, "release_version": "CMSSW_12_1_8"}]

        myInterface = MySQLInterface(logger = logging, engine = None)
        (updatedSQL, bindList) = myInterface.substitute(sql, binds)

        goodSQL = "INSERT INTO RELEASE_VERSIONS (RELEASE_VERSION_ID, RELEASE_VERSION) VALUES (%s, %s)"
        assert goodSQL == updatedSQL, \
               "Error: SQL is different."

        assert len(bindList) == 1, \
               "Error: bindList is malformed."
        assert len(bindList[0]) == 2, \
               "Error: bind list is the wrong length."
        assert bindList[0][0] == 1, \
               "Error: First bind parameter is wrong."
        assert bindList[0][1] == "CMSSW_12_1_8", \
               "Error: Second bind parameter is wrong."

        return

    def testBindSubstitutionB(self):
        """
        _testBindSubstitutionB_

        Test another query that has been causing problems.
        """
        sql = "INSERT INTO FILE_LUMIS (FILE_LUMI_ID, RUN_NUM, LUMI_SECTION_NUM, FILE_ID) VALUES (:file_lumi_id, :run_num, :lumi_section_num, :file_id)"
        binds = [{"lumi_section_num": "27414", "run_num": "1", "file_lumi_id": 1, "file_id": 1},
                 {"lumi_section_num": "26422", "run_num": "1", "file_lumi_id": 2, "file_id": 1},
                 {"lumi_section_num": "29838", "run_num": "1", "file_lumi_id": 3, "file_id": 1}]

        myInterface = MySQLInterface(logger = logging, engine = None)
        (updatedSQL, bindList) = myInterface.substitute(sql, binds)

        goodSQL = "INSERT INTO FILE_LUMIS (FILE_LUMI_ID, RUN_NUM, LUMI_SECTION_NUM, FILE_ID) VALUES (%s, %s, %s, %s)"

        assert updatedSQL == goodSQL, \
               "Error: SQL updated failed."

        assert len(bindList) == 3, \
               "Error: Wrong number of binds."
        assert bindList[0] == (1, '1', '27414', 1), \
               "Error: Bind 0 has wrong values."
        assert bindList[1] == (2, '1', '26422', 1), \
               "Error: Bind 1 has wrong values."
        assert bindList[2] == (3, '1', '29838', 1), \
               "Error: Bind 2 has wrong values."

        return

    def testBindSubstitutionCase(self):
        """
        _testBindSubstitutionCase_

        Verify that bind substitution works correctly when the bind variables in
        the query and the bind variables list have different case.
        """
        sql = "INSERT INTO FILE_LUMIS (FILE_LUMI_ID, RUN_NUM, LUMI_SECTION_NUM, FILE_ID) VALUES (:FILE_LUMI_ID, :RUN_NUM, :LUMI_SECTION_NUM, :FILE_ID)"
        binds = [{"lumi_section_num": "27414", "run_num": "1", "file_lumi_id": 1, "file_id": 1},
                 {"lumi_section_num": "26422", "run_num": "1", "file_lumi_id": 2, "file_id": 1},
                 {"lumi_section_num": "29838", "run_num": "1", "file_lumi_id": 3, "file_id": 1}]

        myInterface = MySQLInterface(logger = logging, engine = None)
        (updatedSQL, bindList) = myInterface.substitute(sql, binds)

        goodSQL = "INSERT INTO FILE_LUMIS (FILE_LUMI_ID, RUN_NUM, LUMI_SECTION_NUM, FILE_ID) VALUES (%s, %s, %s, %s)"

        assert updatedSQL == goodSQL, \
               "Error: SQL updated failed."

        assert len(bindList) == 3, \
               "Error: Wrong number of binds."
        assert bindList[0] == (1, '1', '27414', 1), \
               "Error: Bind 0 has wrong values."
        assert bindList[1] == (2, '1', '26422', 1), \
               "Error: Bind 1 has wrong values."
        assert bindList[2] == (3, '1', '29838', 1), \
               "Error: Bind 2 has wrong values."

        return

if __name__ == "__main__":
    unittest.main()



from __future__ import print_function

import threading
import traceback

import cherrypy


def start(server):
    try:
        server.start(blocking=False)
    except RuntimeError as e:
        # there appears to be worker threads from a previous test
        # hanging out. Try to slay them so that we can keep going
        print("Failed to load cherrypy with exception: %s\n" % e)
        print("The threads are: \n%s\n" % threading.enumerate())
        print("The previous test was %s\n" % server.getLastTest())
        print(traceback.format_exc())
        server.stop()
        raise e


def stop(server):
    server.stop()
    server.setLastTest()
    cherrypy.engine.exit()
    return


#!/usr/bin/env python
import argparse
import getpass
import json
import os
import pwd
import sys
from collections import OrderedDict
from textwrap import TextWrapper

from urllib.error import HTTPError, URLError
from urllib.request import build_opener, HTTPSHandler
from urllib.parse import urlencode, quote_plus
from http.client import HTTPSConnection

# table parameters
SEPARATELINE = "|" + "-" * 51 + "|"
SPLITLINE = "|" + "*" * 51 + "|"

# ID for the User-Agent
CLIENT_ID = 'validate-test-wfs/1.3::python/%s.%s' % sys.version_info[:2]

# Cached DQMGui data
cachedDqmgui = None


class HTTPSClientAuthHandler(HTTPSHandler):
    """
    Basic HTTPS class
    """

    def __init__(self, key, cert):
        HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert

    def https_open(self, req):
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=290):
        return HTTPSConnection(host, key_file=self.key, cert_file=self.cert)


def getX509():
    "Helper function to get x509 from env or tmp file"
    certFile = os.environ.get('X509_USER_CERT', '')
    keyFile = os.environ.get('X509_USER_KEY', '')
    if certFile and keyFile:
        return certFile, keyFile

    proxy = os.environ.get('X509_USER_PROXY', '')
    if not proxy:
        proxy = '/tmp/x509up_u%s' % pwd.getpwuid(os.getuid()).pw_uid
        if not os.path.isfile(proxy):
            return '', ''
    return proxy, proxy


def getContent(url, params=None, headers=None):
    certFile, keyFile = getX509()
    client = '%s (%s)' % (CLIENT_ID, os.environ.get('USER', ''))
    handler = HTTPSClientAuthHandler(keyFile, certFile)
    opener = build_opener(handler)
    if headers:
        opener.addheaders = headers
    else:
        opener.addheaders = [("User-Agent", client),
                             ("Accept", "application/json")]

    try:
        response = opener.open(url, params)
        if "auth/x509" in url:
            output = response.headers.get('X-Rucio-Auth-Token', '')  # option 1
            # output = response.getheader('X-Rucio-Auth-Token')  # option 2
        else:
            output = response.read()
    except HTTPError as e:
        print("The server couldn't fulfill the request at %s" % url)
        print("Error: {}".format(e))
        output = '{}'
        # sys.exit(1)
    except URLError as e:
        print('Failed to reach server at %s' % url)
        print('Reason: ', e.reason)
        sys.exit(2)
    return output


def getRucioToken(rucioUrl):
    """
    Get a Rucio token for this account
    """
    mapHosts = {"http://cms-rucio-int.cern.ch": "https://cms-rucio-auth-int.cern.ch",
                "http://cms-rucio.cern.ch": "https://cms-rucio-auth.cern.ch"}

    rucioAuth = None
    for hostUrl, authUrl in mapHosts.items():
        if hostUrl == rucioUrl:
            rucioAuth = authUrl
            rucioAcct = getpass.getuser()
    if not rucioAuth:
        print("Failed to parse Rucio URL.")
        sys.exit(10)

    rucioAuth = '%s/auth/x509' % rucioAuth
    token = getContent(rucioAuth, headers=[("X-Rucio-Account", rucioAcct)])
    print("Retrieved token for account: {}, token is: {}".format(rucioAcct, token))
    return token


def getDID(rucioUrl, token, dataId):
    """
    Retrieve basic information for a given data identifier (likely a container)
    """
    headers = [("X-Rucio-Auth-Token", token),
               ("Content-type", "application/json"),
               ("Accept", "application/json")]

    rucioUrl = '{}/dids/cms/{}?dynamic=anything'.format(rucioUrl, dataId)
    data = json.loads(getContent(rucioUrl, headers=headers))
    # DEBUG print("Container data retrieved:\n%s" % data)
    return data


def getBlocks(rucioUrl, token, dataId):
    """
    Retrieve information for a CMS block identifier
    """
    headers = [("X-Rucio-Auth-Token", token)]
    # Server doesn't accept these headers below ..
    #           ("Content-type", "application/json"),
    #           ("Accept", "application/json")]
    rucioUrl = '{}/dids/cms/{}/dids'.format(rucioUrl, dataId)
    data = getContent(rucioUrl, headers=headers)
    # DEBUG print("Block data retrieved:\n%s" % data)
    return data


def getBlockMeta(rucioUrl, token, dataId):
    """
    Retrieve the CMS block metadata (like whether the block is opened or not, etc).
    """
    headers = [("X-Rucio-Auth-Token", token),
               ("Content-type", "application/json"),
               ("Accept", "application/json")]

    rucioUrl = '{}/dids/cms/{}/meta'.format(rucioUrl, quote_plus(dataId))
    data = json.loads(getContent(rucioUrl, headers=headers))
    # DEBUG print("Block meta data retrieved:\n%s" % data)
    return data


def getReplicas(rucioUrl, token, dataId):
    """
    Retrieve all the replicas in a given CMS block identifier
    """
    headers = [("X-Rucio-Auth-Token", token)]
    # Here it accepts, but then it returns the wrong response!!!
    #           ("Content-type", "application/json"),
    #           ("Accept", "application/json")]

    rucioUrl = '{}/replicas/cms/{}/datasets'.format(rucioUrl, quote_plus(dataId))
    data = getContent(rucioUrl, headers=headers)
    # DEBUG print("Replicas data retrieved:\n%s" % data)
    return data


def getDIDRules(rucioUrl, token, dataId):
    """
    Retrieve all the rules for a given data identifier
    """
    headers = [("X-Rucio-Auth-Token", token)]
    # Server doesn't accept these headers below ..
    #           ("Content-type", "application/json"),
    #           ("Accept", "application/json")]

    rucioUrl = '{}/dids/cms/{}/rules'.format(rucioUrl, dataId)
    data = getContent(rucioUrl, headers=headers)
    # DEBUG print("getDIDRules data retrieved:\n%s" % data)
    return data


def reader(stream):
    """
    Home-made function to consume newline delimited json streaming data
    """
    for line in stream.split(b"\n"):
        if line:
            yield json.loads(line)


def getCouchSummary(reqName, baseUrl):
    """
    Queries couchdb wmstats database for the workloadSummary
    """
    couchOutput = {}
    couchWorkl = baseUrl + "/couchdb/workloadsummary/" + reqName
    couchOutput["workflow_summary"] = json.loads(getContent(couchWorkl))
    if not couchOutput["workflow_summary"]:
        # then just add the output key to get it going downstream
        couchOutput["workflow_summary"]["output"] = None
    return couchOutput


def getReqMgrCache(reqName, baseUrl):
    """
    Queries Request Manager database for the spec file and a few other stuff
    """
    urn = baseUrl + "/reqmgr2/data/request/" + reqName
    # reqmgr2 returns a different format
    reqmgrOutput = json.loads(getContent(urn))['result'][0][reqName]
    return reqmgrOutput


def getDbsInfo(dataset, baseUrl):
    """
    Queries 3 DBS APIs to get all general information needed
    """
    if 'testbed' in baseUrl:
        dbsUrl = baseUrl + "/dbs/int/global/DBSReader/"
    else:
        dbsUrl = baseUrl + "/dbs/prod/global/DBSReader/"

    dbsOutput = {}
    dbsOutput[dataset] = {"blockorigin": [], "filesummaries": [], "outputconfigs": []}
    for api in dbsOutput[dataset]:
        fullUrl = dbsUrl + api + "?" + urlencode({'dataset': dataset})
        data = json.loads(getContent(fullUrl))
        dbsOutput[dataset][api] = data

    # Separate query for prep_id, since we want any access_type
    fullUrl = dbsUrl + "datasets?" + urlencode({'dataset': dataset})
    fullUrl += "&dataset_access_type=*&detail=True"
    data = json.loads(getContent(fullUrl))
    dbsOutput[dataset]["prep_id"] = data[0]["prep_id"] if data else ''
    return dbsOutput


def getDqmInfo(url):
    """
    Fetch all the data from the DQMGui
    """
    global cachedDqmgui

    dqmguiUrl = url + "/data/json/samples"
    # check whether we have this data in cache
    if cachedDqmgui is None:
        cachedDqmgui = {}
    if url not in cachedDqmgui:
        try:
            cachedDqmgui[url] = json.loads(getContent(dqmguiUrl))
        except Exception:
            print("Failed to fetch data from DQM GUI")
            cachedDqmgui[url]['samples'] = []
    return cachedDqmgui[url]['samples']


def harvesting(workload, outDsets):
    """
    Parse the request spec and query DQMGui in case harvesting is enabled
    """
    if workload['RequestType'] == 'DQMHarvest':
        wantedOutput = [workload['InputDataset']]
    elif str(workload.get('EnableHarvesting', 'False')) == 'True':
        wantedOutput = [dset for dset in outDsets if dset.endswith('/DQMIO') or dset.endswith('/DQM')]
    else:
        return

    urls = workload['DQMUploadUrl'].split(';')
    if not wantedOutput:
        print("Well, it's embarassing! Harvesting is enabled but there is nothing to harvest")
        return

    for url in urls:
        print("Harvesting enabled. Querying DQMGui at: %s" % url)
        allFiles = getDqmInfo(url)
        for outDsets in wantedOutput:
            # dirty hack for multirun harvesting, that changes the dataset name up here
            outDsets = outDsets.rsplit('/', 1)[0]
            for sample in allFiles:
                for item in sample['items']:
                    if item['dataset'].startswith(outDsets):
                        print(item)


def compareLists(d1, d2, d3=None, key=None):
    """
    Receives list of data from different sources and compare them.

    In case a key is passed in, then the input is a dictionary and we validate
    the value of that key from different sources.
    """
    outcome = 'NOPE'
    # just to make comparison easier when the third list is not provided
    if not d3:
        d3 = d2

    try:
        if isinstance(d1, list):
            if len(d1) != len(d2) or len(d1) != len(d3):
                return outcome
            if set(d1) ^ set(d2) or set(d1) ^ set(d3):
                return outcome
        elif isinstance(d1, dict):
            if len(d1) != len(d2) or len(d1) != len(d3):
                return outcome
            for dset in d1:
                if d1[dset][key] != d2[dset][key] or d1[dset][key] != d3[dset][key]:
                    return outcome
        else:
            print("Data type is neither a list nor dict: %s" % type(d1))
            return outcome
    except KeyError:
        pass

    return 'ok'


def compareSpecial(d1, d2, key=None):
    """
    Make special comparisons that requires iterating over all blocks and etc
    """
    if not isinstance(d1, dict):
        print("You must provide a dict data type!")

    outcome = 'NOPE'
    try:
        if key == 'lumis':
            for dset in d2:
                if d1[key] != d2[dset][key]:
                    return outcome

        if key == 'PNN':
            for dset in d2:
                for block, value in d2[dset].items():
                    if isinstance(value, dict):
                        if d1[dset][block][key] != d2[dset][block][key]:
                            return outcome
    except KeyError:
        pass

    return 'ok'


def twClosure(replace_whitespace=False,
              break_long_words=False,
              width=120,
              initial_indent=''):
    """
    Deals with indentation of dictionaries with very long key, value pairs.
    replace_whitespace: Replace each whitespace character with a single space.
    break_long_words: If True words longer than width will be broken.
    width: The maximum length of wrapped lines.
    initial_indent: String that will be prepended to the first line of the output

    Wraps all strings for both keys and values to 120 chars.
    Uses 4 spaces indentation for both keys and values.
    Nested dictionaries and lists go to next line.
    """
    twr = TextWrapper(replace_whitespace=replace_whitespace,
                      break_long_words=break_long_words,
                      width=width,
                      initial_indent=initial_indent)

    def twEnclosed(obj, ind='', reCall=False):
        """
        The inner function of the closure
        ind: Initial indentation for the single output string
        reCall: Flag to indicate a recursive call (should not be used outside)
        """
        output = ''
        if isinstance(obj, dict):
            obj = OrderedDict(sorted(obj.items(),
                                     key=lambda t: t[0],
                                     reverse=False))
            if reCall:
                output += '\n'
            ind += '    '
            for key, value in obj.items():
                output += "%s%s: %s" % (ind,
                                        ''.join(twr.wrap(key)),
                                        twEnclosed(value, ind, reCall=True))
        elif isinstance(obj, list):
            if reCall:
                output += '\n'
            ind += '    '
            for value in obj:
                output += "%s%s" % (ind, twEnclosed(value, ind, reCall=True))
        else:
            output += "%s\n" % str(obj)  # join(twr.wrap(str(obj)))
        return output

    return twEnclosed


def twPrint(obj):
    """
    A simple caller of twClosure (see docstring for twClosure)
    """
    twPrinter = twClosure()
    print(twPrinter(obj))


def handleReqMgr(reqName, reqmgrUrl):
    """
    Query ReqMgr and performs all the processing and dirty keys
    manipulation. In summary:
      1. It gathers the input dataset information
      2. It gathers the list of output datasets according to reqmgr API
      3. If harvesting is enabled, it looks up for the files in the DQM server
      4. It also calculates the workflow runtime (completed - assigned time)

    Returns two dictionaries:
      - (dict) reqmgrInputDset: contains information about the input data
      - (list) reqmgrOutDsets:  contains a list of the output datasets
    """
    reqmgrOut = getReqMgrCache(reqName, reqmgrUrl)

    if reqmgrOut['RequestStatus'] not in ['completed', 'closed-out', 'announced']:
        print("We cannot validate wfs in this state: %s\n" % reqmgrOut['RequestStatus'])
        return (None, None)

    reqmgrInputDset = {}
    try:
        reqmgrInputDset = {'TotalEstimatedJobs': reqmgrOut['TotalEstimatedJobs'],
                           'TotalInputEvents': reqmgrOut['TotalInputEvents'],
                           'TotalInputLumis': reqmgrOut['TotalInputLumis'],
                           'TotalInputFiles': reqmgrOut['TotalInputFiles'],
                           'lumis': reqmgrOut['TotalInputLumis']}  # this lumis is needed for comparison
    except KeyError:
        raise AttributeError("Total* parameters not found in reqmgr_workload_cache database")

    # calculates workflow runtime
    startTime, endTime = 0, 0
    for entry in reqmgrOut['RequestTransition']:
        if entry['Status'] == 'assigned':
            startTime = entry['UpdateTime']
        elif entry['Status'] == 'completed':
            endTime = entry['UpdateTime']
    reqmgrInputDset['Runtime'] = (endTime - startTime) / 3600.  # result in hours

    reqmgrInputDset['InputDataset'] = reqmgrOut['InputDataset'] if 'InputDataset' in reqmgrOut else ''
    if 'Task1' in reqmgrOut and 'InputDataset' in reqmgrOut['Task1']:
        reqmgrInputDset['InputDataset'] = reqmgrOut['Task1']['InputDataset']
    elif 'Step1' in reqmgrOut and 'InputDataset' in reqmgrOut['Step1']:
        reqmgrInputDset['InputDataset'] = reqmgrOut['Step1']['InputDataset']

    reqmgrInputDset['RequestNumEvents'] = reqmgrOut['RequestNumEvents'] if 'RequestNumEvents' in reqmgrOut else ''
    if 'Task1' in reqmgrOut and 'RequestNumEvents' in reqmgrOut['Task1']:
        reqmgrInputDset['RequestNumEvents'] = reqmgrOut['Task1']['RequestNumEvents']
    elif 'Step1' in reqmgrOut and 'RequestNumEvents' in reqmgrOut['Step1']:
        reqmgrInputDset['RequestNumEvents'] = reqmgrOut['Step1']['RequestNumEvents']

    reqmgrOutDsets = reqmgrOut['OutputDatasets']

    # Handle new StepChain/TaskChain output parentage map
    if reqmgrOut['RequestType'] in ('StepChain', 'TaskChain'):
        chainMap = reqmgrOut.get('ChainParentageMap', {})
        if chainMap:
            chainMap = [dset for k, v in chainMap.items() for dset in v['ChildDsets']]
            if set(chainMap) != set(reqmgrOut['OutputDatasets']):
                print("ERROR: list of output datasets doesn't match the chain parentage map")
        else:
            print("WARNING: StepChain/TaskChain workflow without a 'ChainParentageMap' argument!")

    # Handle harvesting case

    print("----------------------------------------------------\nComments:")
    twPrint(reqmgrOut.get('Comments', ''))
    print("----------------------------------------------------\n")

    harvesting(reqmgrOut, reqmgrOutDsets)
    if reqmgrOut['RequestType'] == 'DQMHarvest':
        print("There is nothing else that we can validate here...\n")
        return (None, None)
    return reqmgrInputDset, reqmgrOutDsets


def handleCouch(reqName, reqmgrUrl, reqmgrOutDsets):
    """
    Fetch the workloadSummary directly from couch and perform all
    the processing and dirty keys manipulation.

    Returns a dictionary key by dataset name with their information
    """
    couchOutDsets = {}

    # Get workload summary,  output samples, num of files and events
    couchOutput = getCouchSummary(reqName, reqmgrUrl)
    if couchOutput["workflow_summary"]["output"]:
        for dset in reqmgrOutDsets:
            if dset in couchOutput["workflow_summary"]["output"]:
                couchOutDsets.setdefault(dset, {})
                couchOutDsets[dset]['numFiles'] = couchOutput["workflow_summary"]["output"][dset]['nFiles']
                couchOutDsets[dset]['events'] = couchOutput["workflow_summary"]["output"][dset]['events']
                couchOutDsets[dset]['dsetSize'] = couchOutput["workflow_summary"]["output"][dset]["size"]
            else:
                couchOutDsets[dset] = {'events': '', 'numFiles': '', 'dsetSize': ''}
    return couchOutDsets


def handleRucio(rucioToken, reqmgrOutDsets, rucioUrl):
    """
    Query Rucio and performs all the processing and dirty keys
    manipulation.
    """
    rucioInfo = {}
    if not rucioToken:
        return rucioInfo

    for dset in reqmgrOutDsets:
        rucioInfo.setdefault(dset, {})
        rucioInfo[dset].setdefault('dsetSize', 0)
        rucioInfo[dset].setdefault('numBlocks', 0)
        rucioInfo[dset].setdefault('numFiles', 0)
        containerData = getDID(rucioUrl, rucioToken, dset)
        if not containerData:
            continue
        rucioInfo[dset]['dsetSize'] = containerData['bytes']
        # FIXME should length be the number of blocks?!?
        # rucioInfo[dset]['numBlocks'] = containerData['length']

        rucioInfo[dset].setdefault('rules', [])
        for rule in reader(getDIDRules(rucioUrl, rucioToken, dset)):
            rucioInfo[dset]['rules'].append(rule['id'])

        for block in reader(getBlocks(rucioUrl, rucioToken, containerData['name'])):
            rucioInfo[dset]['numBlocks'] += 1
            blockMeta = getBlockMeta(rucioUrl, rucioToken, block['name'])
            rucioInfo[dset]['numFiles'] += blockMeta['length']

            rucioInfo[dset].setdefault(block['name'], {})
            rucioInfo[dset][block['name']]['numFiles'] = blockMeta['length']
            rucioInfo[dset][block['name']]['is_open'] = blockMeta['is_open']
            rucioInfo[dset][block['name']]['project'] = blockMeta['project']
            for replica in reader(getReplicas(rucioUrl, rucioToken, block['name'])):
                # Look only at the first replica
                rucioInfo[dset][block['name']]['state'] = replica['state']
                rucioInfo[dset][block['name']]['PNN'] = replica['rse'].replace("_Test", "")

    return rucioInfo


def handleDBS(reqmgrOutDsets, cmswebUrl):
    """
    Query DBS and performs all the processing and dirty keys
    manipulation.
    """
    dbsInfo = {}

    for dset in reqmgrOutDsets:
        # Get information from 3 DBS Apis
        dbsOutput = getDbsInfo(dset, cmswebUrl)
        dbsInfo.setdefault(dset, {})
        dbsInfo[dset].setdefault('prep_id', dbsOutput[dset]['prep_id'])
        for item in dbsOutput[dset]['outputconfigs']:
            dbsInfo[dset].setdefault('release', item['release_version'])
        for item in dbsOutput[dset]['filesummaries']:
            dbsInfo[dset].setdefault('dsetSize', item['file_size'])
            dbsInfo[dset].setdefault('numBlocks', item['num_block'])
            dbsInfo[dset].setdefault('numFiles', item['num_file'])
            dbsInfo[dset].setdefault('events', item['num_event'])
            dbsInfo[dset].setdefault('lumis', item['num_lumi'])
        for item in dbsOutput[dset]['blockorigin']:
            dbsInfo[dset].setdefault(item['block_name'], {})
            dbsInfo[dset][item['block_name']] = {'PNN': item['origin_site_name'],
                                                 'numFiles': item['file_count'],
                                                 'isOpen': item['open_for_writing']}
    return dbsInfo


def validateAll(reqmgrInputDset, couchInfo, rucioInfo, dbsInfo):
    """
    Now that we have all the necessary information, we can start performing some
    x-checks. Test cases are:
      1. output dataset name in workload_cache, workloadSummary, rucio and dbs
      2. output dataset size in workloadSummary, rucio and dbs
      3. output number of blocks in rucio and dbs
      4. output number of files in workloadSummary, rucio and dbs
      5. output number of events in workloadSummary and dbs
      6. output number of lumis in the workload_cache and dbs
      7. output PNN in rucio and dbs
      8. whether blocks are closed
    """
    print('' + SPLITLINE)
    print('|' + ' ' * 25 + '| CouchDB | Rucio  | DBS  |')
    print(SEPARATELINE)

    compRes = compareLists(list(couchInfo.keys()), list(rucioInfo.keys()), list(dbsInfo.keys()))
    print('| Same dataset name       | {compRes:7s} | {compRes:6s} | {compRes:4s} |'.format(compRes=compRes))

    compRes = compareLists(couchInfo, rucioInfo, dbsInfo, key='dsetSize')
    print('| Same dataset size       | {compRes:7s} | {compRes:6s} | {compRes:4s} | '.format(compRes=compRes))

    compRes = compareLists(rucioInfo, dbsInfo, key='numBlocks')
    print('| Same number of blocks   | %-7s | %-6s | %-4s |' % ('--', compRes, compRes))

    compRes = compareLists(couchInfo, rucioInfo, dbsInfo, key='numFiles')
    print('| Same number of files    | {compRes:7s} | {compRes:6s} | {compRes:4s} | '.format(compRes=compRes))

    compRes = compareLists(couchInfo, dbsInfo, key='events')
    print('| Same number of events   | %-7s | %-6s | %-4s |' % (compRes, '--', compRes))

    compRes = compareSpecial(reqmgrInputDset, dbsInfo, key='lumis')
    print('| Same number of lumis    | %-7s | %-6s | %-4s |' % (compRes, '--', compRes))

    compRes = compareSpecial(rucioInfo, dbsInfo, key='PNN')
    print('| Same RSE                | %-7s | %-6s | %-4s |' % ('--', compRes, compRes))

    print(SPLITLINE)


def main():
    """
    Requirements: you need to have your proxy and proper x509 environment
     variables set.

    Receive a workflow name in order to fetch the following information:
     - from couchdb: gets the output datasets and spec file
     - from rucio: gets dataset, block, files information
     - from dbs: gets dataset, block and files information
    """
    parser = argparse.ArgumentParser(description="Validate workflow input, output and config")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-w', '--workflow', help='A single workflow name')
    group.add_argument('-i', '--inputFile', help='Plain text file containing request names (one per line)')
    parser.add_argument('-c', '--cms', help='CMSWEB url to talk to DBS. E.g: cmsweb.cern.ch')
    parser.add_argument('-r', '--reqmgr', help='Request Manager URL. Example: couch-dev1.cern.ch')
    parser.add_argument('-v', '--verbose', help='Increase output verbosity', action="store_true")
    parser.add_argument('-x', '--rucio', help='Rucio url', default='cms-rucio-int.cern.ch')
    args = parser.parse_args()

    if args.workflow:
        listRequests = [args.workflow]
    elif args.inputFile:
        with open(args.inputFile, 'r') as f:
            listRequests = [req.rstrip('\n') for req in f.readlines()]
    else:
        parser.error("You must provide either a workflow name or an input file name.")
        sys.exit(3)

    verbose = True if args.verbose else False
    cmswebUrl = "https://" + args.cms if args.cms else "https://cmsweb-testbed.cern.ch"
    reqmgrUrl = "https://" + args.reqmgr if args.reqmgr else "https://cmsweb-testbed.cern.ch"
    rucioUrl = "http://" + args.rucio if args.rucio else "http://cms-rucio-int.cern.ch"

    rucioToken = getRucioToken(rucioUrl)

    for reqName in listRequests:
        print("\n----------------------------------------------------")
        print("==> %s" % reqName)
        # Retrieve and process ReqMgr information
        reqmgrInputDset, reqmgrOutDsets = handleReqMgr(reqName, reqmgrUrl)
        if reqmgrInputDset is None:
            continue

        # Retrieve and process CouchDB information
        couchInfo = handleCouch(reqName, reqmgrUrl, reqmgrOutDsets)

        # Retrieve and process Rucio information
        rucioInfo = handleRucio(rucioToken, reqmgrOutDsets, rucioUrl)

        # Retrieve and process DBS information
        dbsInfo = handleDBS(reqmgrOutDsets, cmswebUrl)

        # Perform all the possible common validations
        validateAll(reqmgrInputDset, couchInfo, rucioInfo, dbsInfo)

        # Starts VERBOSE mode for the information retrieved so far
        if verbose:
            print("\n======> Request information from reqmgr2 db: ")
            twPrint(reqmgrInputDset)
            print("\n======> ReqMgr2 output dataset info: ")
            twPrint(reqmgrOutDsets)
            print("\n======> Couch output dataset info: ")
            twPrint(couchInfo)
            print("\n======> DBS info: ")
            twPrint(dbsInfo)
            print("\n======> Rucio info: ")
            twPrint(rucioInfo)
        print("\n----------------------------------------------------")

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(main())



"""
Unit tests for the WMCore/MicroService/DataStructs/MSOutputTemplate.py module
"""
from __future__ import division, print_function

import unittest
from copy import deepcopy

from builtins import range

from Utils.PythonVersion import PY3

from WMCore.MicroService.MSOutput.MSOutputTemplate import MSOutputTemplate


class MSOutputTemplateTest(unittest.TestCase):
    """
    Test the very basic functionality of the MSOutputTemplate module
    """
    taskchainSpec = {"_id": "taskchain_id",
                     "RequestType": "TaskChain",
                     "SubRequestType": "",
                     "TaskChain": 2,
                     "Campaign": "top-campaign",
                     "RequestName": "taskchain_request_name",
                     "SiteWhitelist": ["Site_1", "Site_2"],
                     "Task1": {"KeepOutput": False,
                               "Campaign": "task1-campaign"},
                     "Task2": {"KeepOutput": False,
                               "Campaign": "task2-campaign"},
                     "OutputDatasets": ["output-dataset-1", "output-dataset-2"],
                     "ChainParentageMap": {"Task1": {"ParentDset": None,
                                                     "ChildDsets": ["output-dataset-1"]},
                                           "Task2": {"ParentDset": "output-dataset-1",
                                                     "ChildDsets": ["output-dataset-2"]}}
                     }
    stepchainSpec = {"_id": "stepchain_id",
                     "RequestType": "StepChain",
                     "SubRequestType": "",
                     "StepChain": 2,
                     "Campaign": "top-campaign",
                     "RequestName": "stepchain_request_name",
                     "SiteWhitelist": ["Site_1", "Site_2"],
                     "Step1": {"KeepOutput": False,
                               "Campaign": "step1-campaign"},
                     "Step2": {"KeepOutput": False,
                               "Campaign": "step2-campaign"},
                     "OutputDatasets": ["output-dataset-1", "output-dataset-2"],
                     "ChainParentageMap": {"Step1": {"ParentDset": None,
                                                     "ChildDsets": ["output-dataset-1"]},
                                           "Step2": {"ParentDset": "output-dataset-1",
                                                     "ChildDsets": ["output-dataset-2"]}}
                     }
    rerecoSpec = {"_id": "rereco_id",
                  "RequestType": "ReReco",
                  "SubRequestType": "",
                  "Campaign": "top-campaign",
                  "RequestName": "rereco_request_name",
                  "SiteWhitelist": ["Site_1", "Site_2"],
                  "OutputDatasets": ["output-dataset-1", "output-dataset-2"]
                  }

    mongoDoc = {"_id": "any-mongo-doc-id",
                "RequestName": "mongo-doc-name",
                "RequestType": "mongo-doc-type",
                "Campaign": "top-campaign",
                "CreationTime": 123,
                "LastUpdate": 123456,
                "IsRelVal": False,
                "OutputDatasets": ["output-dataset-1", "output-dataset-2"],
                "OutputMap": [{'Campaign': 'campaign-1',
                               'Dataset': 'output-dataset-1',
                               'DatasetSize': 123,
                               'Copies': 1,
                               'DiskDestination': "",
                               'TapeDestination': "",
                               'DiskRuleID': "",
                               'TapeRuleID': ""},
                              {'Campaign': 'campaign-2',
                               'Dataset': 'output-dataset-2',
                               'DatasetSize': 456,
                               'Copies': 0,
                               'DiskDestination': "",
                               'TapeDestination': "",
                               'DiskRuleID': "",
                               'TapeRuleID': ""}],
                "TransferStatus": "pending"
                }

    outputMapKeys = ["Campaign", "Copies", "Dataset", "DatasetSize", "DiskDestination",
                     "DiskRuleID", "TapeDestination", "TapeRuleID"]

    def setUp(self):
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def testTaskChainSpec(self):
        """
        Test creating a MSOutputTemplate object out of a TaskChain request dictionary
        """
        msOutDoc = MSOutputTemplate(self.taskchainSpec)

        self.assertTrue(msOutDoc["CreationTime"] > 0)
        self.assertIsNone(msOutDoc["LastUpdate"])
        self.assertFalse(msOutDoc["IsRelVal"])
        self.assertEqual(msOutDoc["TransferStatus"], "pending")
        self.assertEqual(msOutDoc["Campaign"], self.taskchainSpec["Campaign"])
        self.assertEqual(msOutDoc["OutputDatasets"], self.taskchainSpec["OutputDatasets"])
        self.assertEqual(msOutDoc["RequestName"], self.taskchainSpec["RequestName"])
        self.assertEqual(msOutDoc["RequestType"], self.taskchainSpec["RequestType"])
        self.assertEqual(msOutDoc["_id"], self.taskchainSpec["_id"])
        self.assertEqual(len(msOutDoc["OutputMap"]), 2)
        self.assertItemsEqual(list(msOutDoc["OutputMap"][0]), self.outputMapKeys)
        for idx in range(2):
            self.assertEqual(msOutDoc["OutputMap"][idx]["DatasetSize"], 0)
            if msOutDoc["OutputMap"][idx]["Campaign"] == self.taskchainSpec["Task1"]["Campaign"]:
                self.assertEqual(msOutDoc["OutputMap"][idx]["Dataset"], "output-dataset-1")
            else:
                self.assertEqual(msOutDoc["OutputMap"][idx]["Campaign"],
                                 self.taskchainSpec["Task2"]["Campaign"])
                self.assertEqual(msOutDoc["OutputMap"][idx]["Dataset"], "output-dataset-2")
            self.assertEqual(msOutDoc["OutputMap"][idx]["Copies"], 1)
            self.assertEqual(msOutDoc["OutputMap"][idx]["DiskDestination"], "")
            self.assertEqual(msOutDoc["OutputMap"][idx]["DiskRuleID"], "")
            self.assertEqual(msOutDoc["OutputMap"][idx]["TapeDestination"], "")
            self.assertEqual(msOutDoc["OutputMap"][idx]["TapeRuleID"], "")

    def testStepChainSpec(self):
        """
        Test creating a MSOutputTemplate object out of a StepChain request dictionary
        """
        msOutDoc = MSOutputTemplate(self.stepchainSpec)

        self.assertTrue(msOutDoc["CreationTime"] > 0)
        self.assertIsNone(msOutDoc["LastUpdate"])
        self.assertFalse(msOutDoc["IsRelVal"])
        self.assertEqual(msOutDoc["TransferStatus"], "pending")
        self.assertEqual(msOutDoc["Campaign"], self.stepchainSpec["Campaign"])
        self.assertEqual(msOutDoc["OutputDatasets"], self.stepchainSpec["OutputDatasets"])
        self.assertEqual(msOutDoc["RequestName"], self.stepchainSpec["RequestName"])
        self.assertEqual(msOutDoc["RequestType"], self.stepchainSpec["RequestType"])
        self.assertEqual(msOutDoc["_id"], self.stepchainSpec["_id"])
        self.assertEqual(len(msOutDoc["OutputMap"]), 2)
        self.assertItemsEqual(list(msOutDoc["OutputMap"][0]), self.outputMapKeys)
        for idx in range(2):
            self.assertEqual(msOutDoc["OutputMap"][idx]["DatasetSize"], 0)
            if msOutDoc["OutputMap"][idx]["Campaign"] == self.stepchainSpec["Step1"]["Campaign"]:
                self.assertEqual(msOutDoc["OutputMap"][idx]["Dataset"], "output-dataset-1")
            else:
                self.assertEqual(msOutDoc["OutputMap"][idx]["Campaign"],
                                 self.stepchainSpec["Step2"]["Campaign"])
                self.assertEqual(msOutDoc["OutputMap"][idx]["Dataset"], "output-dataset-2")

    def testReRecoSpec(self):
        """
        Test creating a MSOutputTemplate object out of a ReReco request dictionary
        """
        msOutDoc = MSOutputTemplate(self.rerecoSpec)

        self.assertTrue(msOutDoc["CreationTime"] > 0)
        self.assertIsNone(msOutDoc["LastUpdate"])
        self.assertFalse(msOutDoc["IsRelVal"])
        self.assertEqual(msOutDoc["TransferStatus"], "pending")
        self.assertEqual(msOutDoc["Campaign"], self.rerecoSpec["Campaign"])
        self.assertEqual(msOutDoc["OutputDatasets"], self.rerecoSpec["OutputDatasets"])
        self.assertEqual(msOutDoc["RequestName"], self.rerecoSpec["RequestName"])
        self.assertEqual(msOutDoc["RequestType"], self.rerecoSpec["RequestType"])
        self.assertEqual(msOutDoc["_id"], self.rerecoSpec["_id"])
        self.assertEqual(len(msOutDoc["OutputMap"]), 2)
        self.assertEqual(msOutDoc["OutputMap"][0]["DatasetSize"], 0)
        self.assertItemsEqual(list(msOutDoc["OutputMap"][0]), self.outputMapKeys)
        self.assertEqual(msOutDoc["OutputMap"][0]["Campaign"], self.rerecoSpec["Campaign"])
        self.assertEqual(msOutDoc["OutputMap"][1]["Campaign"], self.rerecoSpec["Campaign"])
        self.assertItemsEqual([msOutDoc["OutputMap"][0]["Dataset"], msOutDoc["OutputMap"][1]["Dataset"]],
                              self.rerecoSpec["OutputDatasets"])

    def testMongoDoc(self):
        """
        Test creating a MSOutputTemplate object out of a MongoDB record
        """
        msOutDoc = MSOutputTemplate(self.mongoDoc, producerDoc=False)
        self.assertFalse(msOutDoc["IsRelVal"])

        self.assertEqual(msOutDoc["TransferStatus"], "pending")
        self.assertEqual(msOutDoc["CreationTime"], self.mongoDoc["CreationTime"])
        self.assertEqual(msOutDoc["LastUpdate"], self.mongoDoc["LastUpdate"])

        self.assertEqual(msOutDoc["Campaign"], self.mongoDoc["Campaign"])
        self.assertEqual(msOutDoc["OutputDatasets"], self.mongoDoc["OutputDatasets"])
        self.assertEqual(msOutDoc["RequestName"], self.mongoDoc["RequestName"])
        self.assertEqual(msOutDoc["RequestType"], self.mongoDoc["RequestType"])
        self.assertEqual(msOutDoc["_id"], self.mongoDoc["_id"])
        self.assertEqual(len(msOutDoc["OutputMap"]), 2)
        self.assertItemsEqual([msOutDoc["OutputMap"][0]["Dataset"], msOutDoc["OutputMap"][1]["Dataset"]],
                              self.mongoDoc["OutputDatasets"])
        self.assertTrue(msOutDoc["OutputMap"][0]["DatasetSize"] in [123, 456])

        newDoc = deepcopy(self.mongoDoc)
        newDoc.update({"IsRelVal": True, "TransferStatus": "done", "LastUpdate": 333})
        msOutDoc = MSOutputTemplate(newDoc, producerDoc=False)
        self.assertTrue(msOutDoc["IsRelVal"])
        self.assertEqual(msOutDoc["TransferStatus"], "done")
        self.assertEqual(msOutDoc["LastUpdate"], 333)

    def testSetters(self):
        """
        Test the MSOutputTemplate setter methods
        """
        msOutDoc = MSOutputTemplate(self.mongoDoc)
        self.assertFalse(msOutDoc["IsRelVal"])
        msOutDoc._setRelVal({"SubRequestType": "RelVal"})
        msOutDoc.setKey("IsRelVal", True)
        self.assertTrue(msOutDoc["IsRelVal"])

        with self.assertRaises(KeyError):
            msOutDoc.setKey("alan", True)

        msOutDoc.updateDoc({"IsRelVal": False, "LastUpdate": 444})
        self.assertFalse(msOutDoc["IsRelVal"])
        self.assertEqual(msOutDoc["LastUpdate"], 444)

        msOutDoc.updateTime()
        self.assertTrue(msOutDoc["LastUpdate"] > 444)

        with self.assertRaises(RuntimeError):
            msOutDoc.setTransferStatus("bad_status")
        self.assertEqual(msOutDoc["TransferStatus"], "pending")
        msOutDoc.setTransferStatus("done")
        self.assertEqual(msOutDoc["TransferStatus"], "done")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_LumiBased_t

Test lumi based splitting.
"""

from __future__ import division

from builtins import range
import threading
import unittest
import random

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.DataStructs.Run import Run
from WMCore.DAOFactory import DAOFactory
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit
from WMCore.ACDC.DataCollectionService import DataCollectionService
from WMCore.WMSpec.WMWorkload import newWorkload


def getJob(workload):
    """
    getJob

    Given a workload, get a job from it
    """
    job = Job()
    job["task"] = workload.getTask("reco").getPathName()
    job["workflow"] = workload.name()
    job["location"] = "T1_US_FNAL"
    job["owner"] = "evansde77"
    job["group"] = "DMWM"
    return job


class LumiBasedTest(unittest.TestCase):
    """
    _LumiBasedTest_

    Test job splitting using LumiBased
    """

    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        # self.testInit.clearDatabase(modules = ['WMCore.WMBS'])
        self.testInit.setSchema(customModules=["WMCore.WMBS"], useDefault=False)
        self.testInit.setupCouch("lumi_t", "GroupUser", "ACDC")

        myThread = threading.currentThread()
        daofactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger, dbinterface=myThread.dbi)

        locationAction = daofactory(classname = "Locations.New")
        locationAction.execute(siteName='s1', pnn="T1_US_FNAL_Disk")
        locationAction.execute(siteName='s2', pnn="T2_CH_CERN")

        self.testWorkflow = Workflow(spec="spec.xml", owner="mnorman", name="wf001", task="Test")
        self.testWorkflow.create()

        self.performanceParams = {'timePerEvent': 12,
                                  'memoryRequirement': 2300,
                                  'sizePerEvent': 400}

        return

    def tearDown(self):
        """
        _tearDown_

        """
        self.testInit.clearDatabase()
        self.testInit.tearDownCouch()
        return

    def createSubscription(self, nFiles, lumisPerFile, twoSites=False, rand=False):
        """
        _createSubscription_

        Create a subscription for testing
        """

        baseName = makeUUID()

        testFileset = Fileset(name=baseName)
        testFileset.create()
        parentFile = File('%s_parent' % baseName, size=1000, events=100,
                          locations=set(["T1_US_FNAL_Disk"]))
        parentFile.create()
        for i in range(nFiles):
            newFile = File(lfn='%s_%i' % (baseName, i), size=1000,
                           events=100, locations="T1_US_FNAL_Disk")
            lumis = []
            for lumi in range(lumisPerFile):
                if rand:
                    lumis.append(random.randint(1000 * i, 1000 * (i + 1)))
                else:
                    lumis.append((100 * i) + lumi)
            newFile.addRun(Run(i, *lumis))
            newFile.create()
            newFile.addParent(parentFile['lfn'])
            testFileset.addFile(newFile)
        if twoSites:
            for i in range(nFiles):
                newFile = File(lfn='%s_%i_2' % (baseName, i), size=1000,
                               events=100, locations="T2_CH_CERN")
                lumis = []
                for lumi in range(lumisPerFile):
                    if rand:
                        lumis.append(random.randint(1000 * i, 1000 * (i + 1)))
                    else:
                        lumis.append((100 * i) + lumi)
                newFile.addRun(Run(i, *lumis))
                newFile.create()
                newFile.addParent(parentFile['lfn'])
                testFileset.addFile(newFile)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset, workflow=self.testWorkflow,
                                        split_algo="LumiBased", type="Processing")
        testSubscription.create()

        return testSubscription

    def testA_FileSplitting(self):
        """
        _FileSplitting_

        Test that things work if we split files between jobs
        """
        splitter = SplitterFactory()

        oneSetSubscription = self.createSubscription(nFiles=10, lumisPerFile=1)
        jobFactory = splitter(package="WMCore.WMBS", subscription=oneSetSubscription)

        jobGroups = jobFactory(lumis_per_job=3, halt_job_on_file_boundaries=True, performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertTrue(len(job['input_files']), 1)
            self.assertEqual(job['estimatedJobTime'], 100 * 12)
            self.assertEqual(job['estimatedDiskUsage'], 100 * 400)
            self.assertEqual(job['estimatedMemoryUsage'], 2300)

        twoLumiFiles = self.createSubscription(nFiles=5, lumisPerFile=2)
        jobFactory = splitter(package="WMCore.WMBS", subscription=twoLumiFiles)
        jobGroups = jobFactory(lumis_per_job=1, halt_job_on_file_boundaries=True, performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertEqual(len(job['input_files']), 1)
            self.assertEqual(job['estimatedJobTime'], 50 * 12)
            self.assertEqual(job['estimatedDiskUsage'], 50 * 400)
            self.assertEqual(job['estimatedMemoryUsage'], 2300)

        wholeLumiFiles = self.createSubscription(nFiles=5, lumisPerFile=3)
        jobFactory = splitter(package="WMCore.WMBS", subscription=wholeLumiFiles)
        jobGroups = jobFactory(lumis_per_job=2, halt_job_on_file_boundaries=True, performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        # 10 because we split on run boundaries
        self.assertEqual(len(jobGroups[0].jobs), 10)
        jobList = jobGroups[0].jobs
        for idx, job in enumerate(jobList, start=1):
            # Have should have one file, half two
            self.assertEqual(len(job['input_files']), 1)
            if idx % 2 == 0:
                self.assertEqual(job['estimatedJobTime'], (1.0 * round(100 / 3)) * 12)
                self.assertEqual(job['estimatedDiskUsage'], (1.0 * round(100 / 3)) * 400)
            else:
                self.assertEqual(job['estimatedJobTime'], (2.0 * round(100 / 3)) * 12)
                self.assertEqual(job['estimatedDiskUsage'], (2.0 * round(100 / 3)) * 400)
            self.assertEqual(job['estimatedMemoryUsage'], 2300)

        mask0 = jobList[0]['mask'].getRunAndLumis()
        self.assertEqual(mask0, {0: [[0, 1]]})
        mask1 = jobList[1]['mask'].getRunAndLumis()
        self.assertEqual(mask1, {0: [[2, 2]]})
        mask2 = jobList[2]['mask'].getRunAndLumis()
        self.assertEqual(mask2, {1: [[100, 101]]})
        mask3 = jobList[3]['mask'].getRunAndLumis()
        self.assertEqual(mask3, {1: [[102, 102]]})

        j0 = Job(id=jobList[0]['id'])
        j0.loadData()
        self.assertEqual(j0['mask'].getRunAndLumis(), {0: [[0, 1]]})

        # Do it with multiple sites
        twoSiteSubscription = self.createSubscription(nFiles=5, lumisPerFile=2, twoSites=True)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=twoSiteSubscription)
        jobGroups = jobFactory(lumis_per_job=1,
                               halt_job_on_file_boundaries=True,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 2)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertEqual(len(job['input_files']), 1)
            self.assertEqual(job['estimatedJobTime'], 50 * 12)
            self.assertEqual(job['estimatedDiskUsage'], 50 * 400)
            self.assertEqual(job['estimatedMemoryUsage'], 2300)

    def testB_NoRunNoFileSplitting(self):
        """
        _NoRunNoFileSplitting_

        Test the splitting algorithm in the odder fringe
        cases that might be required.
        """
        splitter = SplitterFactory()
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)

        jobGroups = jobFactory(lumis_per_job=3,
                               halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 9)

        # The first job should have three lumis from one run
        # The second three lumis from two different runs
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        self.assertEqual(jobs[0]['estimatedJobTime'], 60 * 12)
        self.assertEqual(jobs[0]['estimatedDiskUsage'], 60 * 400)
        self.assertEqual(jobs[0]['estimatedMemoryUsage'], 2300)
        job1runLumi = jobs[1]['mask'].getRunAndLumis()
        self.assertEqual(job1runLumi[0][0][0] + 1, job1runLumi[0][0][1])  # Run 0, startLumi+1 == endLumi
        self.assertEqual(job1runLumi[1][0][0], job1runLumi[1][0][1])  # Run 1, startLumi == endLumi
        self.assertEqual(jobs[1]['estimatedJobTime'], 60 * 12)
        self.assertEqual(jobs[1]['estimatedDiskUsage'], 60 * 400)
        self.assertEqual(jobs[1]['estimatedMemoryUsage'], 2300)


        # And it should still be the same when you load it out of the database
        j1 = Job(id=jobs[1]['id'])
        j1.loadData()
        self.assertEqual(j1['mask'].getRunAndLumis(), {0: [[3, 4]], 1: [[100, 100]]})

        # Assert that this works differently with file splitting on and run splitting on
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(lumis_per_job=3,
                               halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 10)

        # In this case it should slice things up so that each job only has one run
        # in it.
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {0: [[3, 4]]})

        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=4, twoSites=False)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(lumis_per_job=10,
                               halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0]['mask']['runAndLumis'], {0: [[0, 3]], 1: [[100, 103]], 2: [[200, 201]]})
        self.assertEqual(jobs[1]['mask']['runAndLumis'], {2: [[202, 203]], 3: [[300, 303]], 4: [[400, 403]]})

        j = Job(id=jobs[0]['id'])
        j.loadData()
        self.assertEqual(len(j['input_files']), 3)
        for f in j['input_files']:
            self.assertTrue(f['events'], 100)
            self.assertTrue(f['size'], 1000)
        return

    def createTestWorkload(self):
        """
        _createTestWorkload_

        """
        workload = newWorkload("ACDCTest")
        reco = workload.newTask("reco")
        skim1 = reco.addTask("skim1")
        workload.setOwnerDetails(name="evansde77", group="DMWM")

        # first task uses the input dataset
        reco.addInputDataset(name='/PRIMARY/processed-v1/TIERONE',
                             primary="PRIMARY", processed="processed-v1", tier="TIERONE")
        cmsRunReco = reco.makeStep("cmsRun1")
        cmsRunReco.setStepType("CMSSW")
        reco.applyTemplates()
        cmsRunRecoHelper = cmsRunReco.getTypeHelper()
        cmsRunRecoHelper.addOutputModule("outputRECO",
                                         primaryDataset="PRIMARY",
                                         processedDataset="processed-v2",
                                         dataTier="TIERTWO",
                                         lfnBase="/store/dunkindonuts",
                                         mergedLFNBase="/store/kfc")
        # second step uses an input reference
        cmsRunSkim = skim1.makeStep("cmsRun2")
        cmsRunSkim.setStepType("CMSSW")
        skim1.applyTemplates()
        skim1.setInputReference(cmsRunReco, outputModule="outputRECO")

        return workload

    def testC_ACDCTest(self):
        """
        _ACDCTest_

        Test whether we can get a goodRunList out of ACDC
        and process it correctly.
        """
        workload = self.createTestWorkload()
        dcs = DataCollectionService(url=self.testInit.couchUrl, database=self.testInit.couchDbName)

        testFileA = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileA.addRun(Run(1, 1, 2))
        testFileA.create()
        testFileB = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileB.addRun(Run(1, 3))
        testFileB.create()
        testJobA = getJob(workload)
        testJobA.addFile(testFileA)
        testJobA.addFile(testFileB)

        testFileC = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileC.addRun(Run(1, 4, 6))
        testFileC.create()
        testJobB = getJob(workload)
        testJobB.addFile(testFileC)

        testFileD = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileD.addRun(Run(1, 7))
        testFileD.create()
        testJobC = getJob(workload)
        testJobC.addFile(testFileD)

        testFileE = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileE.addRun(Run(1, 11, 12))
        testFileE.create()
        testJobD = getJob(workload)
        testJobD.addFile(testFileE)

        testFileF = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileF.addRun(Run(2, 5, 6, 7))
        testFileF.create()
        testJobE = getJob(workload)
        testJobE.addFile(testFileF)

        testFileG = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileG.addRun(Run(2, 10, 11, 12))
        testFileG.create()
        testJobF = getJob(workload)
        testJobF.addFile(testFileG)

        testFileH = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileH.addRun(Run(2, 15))
        testFileH.create()
        testJobG = getJob(workload)
        testJobG.addFile(testFileH)

        testFileI = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileI.addRun(Run(3, 20))
        testFileI.create()
        testJobH = getJob(workload)
        testJobH.addFile(testFileI)

        testFileJ = File(lfn=makeUUID(), size=1024, events=1024, locations="T1_US_FNAL_Disk")
        testFileJ.addRun(Run(1, 9))
        testFileJ.create()
        testJobI = getJob(workload)
        testJobI.addFile(testFileJ)

        # dcs.failedJobs([testJobA, testJobB, testJobC, testJobD, testJobE,
        #                testJobF, testJobG, testJobH, testJobI])

        dcs.failedJobs([testJobA, testJobD, testJobH])

        baseName = makeUUID()

        testFileset = Fileset(name=baseName)
        testFileset.create()
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.addFile(testFileD)
        testFileset.addFile(testFileE)
        testFileset.addFile(testFileF)
        testFileset.addFile(testFileG)
        testFileset.addFile(testFileH)
        testFileset.addFile(testFileI)
        testFileset.addFile(testFileJ)
        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="LumiBased",
                                        type="Processing")
        testSubscription.create()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)

        jobGroups = jobFactory(lumis_per_job=100,
                               halt_job_on_file_boundaries=False,
                               splitOnRun=True,
                               collectionName=workload.name(),
                               filesetName=workload.getTask("reco").getPathName(),
                               owner="evansde77",
                               group="DMWM",
                               couchURL=self.testInit.couchUrl,
                               couchDB=self.testInit.couchDbName,
                               performance=self.performanceParams)

        self.assertEqual(jobGroups[0].jobs[0]['mask'].getRunAndLumis(), {1: [[1, 2], [3, 3], [11, 12]]})
        self.assertEqual(jobGroups[0].jobs[1]['mask'].getRunAndLumis(), {3: [[20, 20]]})

        return

    def testD_NonContinuousLumis(self):
        """
        _NonContinuousLumis_

        Test and see if LumiBased can work when the lumis are non continuous
        """

        baseName = makeUUID()
        nFiles = 10

        testFileset = Fileset(name=baseName)
        testFileset.create()
        for i in range(nFiles):
            newFile = File(lfn='%s_%i' % (baseName, i), size=1000,
                           events=100, locations="T1_US_FNAL_Disk")
            # Set to two non-continuous lumi numbers
            lumis = [100 + i, 200 + i]
            newFile.addRun(Run(i, *lumis))
            newFile.create()
            testFileset.addFile(newFile)

        testFileset.commit()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="LumiBased",
                                        type="Processing")
        testSubscription.create()

        splitter = SplitterFactory()
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)

        jobGroups = jobFactory(lumis_per_job=2,
                               halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 10)
        for j in jobs:
            runs = j['mask'].getRunAndLumis()
            for r in runs:
                self.assertEqual(len(runs[r]), 2)
                for l in runs[r]:
                    # Each run should have two lumis
                    # Each lumi should be of form [x, x]
                    # meaning that the first and last lumis are the same
                    self.assertEqual(len(l), 2)
                    self.assertEqual(l[0], l[1])
            self.assertEqual(j['estimatedJobTime'], 100 * 12)
            self.assertEqual(j['estimatedDiskUsage'], 100 * 400)
            self.assertEqual(j['estimatedMemoryUsage'], 2300)

        return

    def testE_getParents(self):
        """
        _getParents_

        Test the TwoFileBased version of this code
        """

        splitter = SplitterFactory()

        oneSetSubscription = self.createSubscription(nFiles=10, lumisPerFile=1)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=oneSetSubscription)

        jobGroups = jobFactory(lumis_per_job=3,
                               split_files_between_job=True,
                               include_parents=True,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertTrue(len(job['input_files']), 1)
            f = job['input_files'][0]
            self.assertEqual(len(f['parents']), 1)
            self.assertEqual(f['lfn'].split('_')[0],
                             list(f['parents'])[0]['lfn'].split('_')[0])

        return

    def testF_RunWhitelist(self):
        """
        _runWhitelist_

        Apparently we're too stupid to do the runlist in
        the GoodRunlist where it would make sense.
        """

        splitter = SplitterFactory()

        oneSetSubscription = self.createSubscription(nFiles=10, lumisPerFile=1)
        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=oneSetSubscription)

        jobGroups = jobFactory(lumis_per_job=10,
                               split_files_between_job=True,
                               runWhitelist=[1],
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 1)
        self.assertEqual(len(jobGroups[0].jobs[0]['input_files']), 1)
        self.assertEqual(len(jobGroups[0].jobs[0]['input_files'][0]['runs']), 1)
        self.assertEqual(jobGroups[0].jobs[0]['input_files'][0]['runs'][0].run, 1)
        return

    def test_NotEnoughEvents(self):
        """
        _test_NotEnoughEvents_

        Checks whether jobs are not created when there are not enough files (actually, events)
        according to the events_per_job requested to the splitter algorithm
        """
        splitter = SplitterFactory()

        # Very small fileset (single file) without enough events
        testSubscription = self.createSubscription(nFiles=1, lumisPerFile=2)

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(lumis_per_job=5,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 0)

        # Still a small fileset (two files) without enough events
        testSubscription = self.createSubscription(nFiles=2, lumisPerFile=2)

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(lumis_per_job=5,
                               performance=self.performanceParams,
                               splitOnRun=False)

        self.assertEqual(len(jobGroups), 0)

        # Finally an acceptable fileset size (three files) with enough events
        testSubscription = self.createSubscription(nFiles=3, lumisPerFile=2)

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(lumis_per_job=5,
                               performance=self.performanceParams,
                               splitOnRun=False)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 2)
        self.assertEqual(len(jobs[0]['input_files']), 3)
        self.assertEqual(len(jobs[1]['input_files']), 1)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 1]], 1: [[100, 101]], 2: [[200, 200]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {2: [[201, 201]]})

        # Test fileset with a single run and splitOnRun=True
        testFileset = Fileset(name="FilesetA")
        for i in range(3):
            testFile = File(lfn="/this/is/file%s" % i, size=1024, events=200, locations="T1_US_FNAL_Disk")
            lumis = [i *  2 + val for val in range(1, 3)]
            testFile.addRun(Run(1, lumis))
            testFile.create()
            testFileset.addFile(testFile)
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="LumiBased",
                                        type="Processing")
        testSubscription.create()

        jobFactory = splitter(package="WMCore.WMBS",
                              subscription=testSubscription)
        jobGroups = jobFactory(lumis_per_job=5,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 2)
        self.assertEqual(len(jobs[0]['input_files']), 3)
        self.assertEqual(len(jobs[1]['input_files']), 1)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {1: [[1, 2], [3, 4], [5, 5]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {1: [[6, 6]]})

        return


if __name__ == '__main__':
    unittest.main()



"""
Unit tests for Unified/TaskManager.py module

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""
from __future__ import division, print_function

import time
import unittest

from builtins import str as newstr
from future.utils import listvalues

from WMCore.MicroService.MSCore.TaskManager import \
    TaskManager, genkey, UidSet

def myFunc(interval, results):
    "Test function"
    time.sleep(interval)
    results.update({interval: 'ok_%s' % interval})


class TaskManagerTest(unittest.TestCase):
    "Unit test for TaskManager module"

    def setUp(self):
        pass

    def testUidSet(self):
        "Test UidSet class"
        mgr = UidSet()
        mgr.add(1)
        mgr.add(1)
        self.assertEqual(2, mgr.get(1))
        mgr.discard(1)
        self.assertEqual(1, mgr.get(1))
        mgr.discard(1)
        self.assertEqual(0, mgr.get(1))

    def testTaskManager(self):
        "Test TaskManager class"
        mgr = TaskManager(nworkers=3)
        self.assertEqual(3, mgr.nworkers())

        jobs = []
        results = {}
        kwds = {}
        args1 = (1, results)
        args2 = (2, results)
        args3 = (3, results)
        args4 = (4, results)
        pid1 = genkey(str(args1) + str(kwds))
        pid2 = genkey(str(args2) + str(kwds))
        pid3 = genkey(str(args3) + str(kwds))
        pid4 = genkey(newstr(args4) + newstr(kwds))

        jobs.append(mgr.spawn(myFunc, *args1))
        jobs.append(mgr.spawn(myFunc, *args2))
        jobs.append(mgr.spawn(myFunc, *args3))
        jobs.append(mgr.spawn(myFunc, *args4))

        self.assertEqual(True, mgr.is_alive(pid1))
        self.assertEqual(True, mgr.is_alive(pid2))
        self.assertEqual(True, mgr.is_alive(pid3))
        self.assertEqual(True, mgr.is_alive(pid4))

        mgr.joinall(jobs)

        self.assertEqual(sorted(results.keys()), [1, 2, 3, 4])
        self.assertEqual(listvalues(results), ['ok_1', 'ok_2', 'ok_3', 'ok_4'])

        self.assertEqual(False, mgr.is_alive(pid1))
        self.assertEqual(False, mgr.is_alive(pid2))
        self.assertEqual(False, mgr.is_alive(pid3))

        for worker in mgr.workers:
            self.assertEqual(True, worker.is_alive())

        print("### TaskManager quit")
        mgr.quit()
        print("### TaskManager workrers no longer running")

        for worker in mgr.workers:
            self.assertEqual(False, worker.is_alive())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_LCGImpl_

Implementation of StageOutImpl interface for lcg-cp

"""
import os

from future.utils import viewitems

from WMCore.Storage.Registry import registerStageOutImpl
from WMCore.Storage.StageOutImpl import StageOutImpl

_CheckExitCodeOption = True


class LCGImpl(StageOutImpl):
    """
    _LCGImpl_

    Implement interface for srmcp v2 command with lcg-* commands

    """

    def __init__(self, stagein=False):
        StageOutImpl.__init__(self, stagein)

        self.setups = ''
        self.timeoutOptions = '--srm-timeout 600 --sendreceive-timeout 600 --connect-timeout 300'
        setupScripts = {
            'OSG_GRID':           '/setup.sh',
            'GLITE_WMS_LOCATION': '/etc/profile.d/glite-wmsui.sh',
            'GLITE_LOCATION':     '/../etc/profile.d/grid-env.sh',
            'GRID_ENV_LOCATION':  '/grid-env.sh',
        }

        for env, script in viewitems(setupScripts):
            if os.environ.get(env):
                fullScript = os.path.normpath(os.path.join(os.environ[env], script))
                if os.path.isfile(fullScript):
                    self.setups += 'source %s; ' % fullScript
        self.setups += 'date "+%Y-%m-%dT%H:%M:%S"; '

    def createSourceName(self, protocol, pfn):
        """
        _createSourceName_

        SRM uses file:/// urls

        """
        if pfn.startswith('/'):
            return "file:%s" % pfn
        elif os.path.isfile(pfn):
            return "file:%s" % os.path.abspath(pfn)
        else:
            return pfn

    def createOutputDirectory(self, targetPFN):
        """
        _createOutputDirectory_

        lcg-cp handles this

        """
        return

    def createRemoveFileCommand(self, pfn):
        """
        handle both srm and file pfn types

        lcg-del options used:
          -b don't make BDII calls to get SE type. SURLs must be fully provided
          -l it means that the SURL is not registered in any LFC server, don't connect to the LFC server
          -D specifies the default SE type we want to use
          --srm-timeout         Sets  the SRM timeout, define the maximum waiting time for a SRM query.
                                The request will be aborted if it is still queued after srmtimeout seconds.
                                Default to 180 seconds.
          --sendreceive-timeout Sets the send/receive timeout, maximum total time for the operation.
                                The operation will be aborted if the operation is not finished before X seconds.
                                Default to 3600 seconds.
          --connect-timeout     The connection will be aborted if the remote host doesn't
                                reply before X seconds. Default to 60 seconds.
        """
        if pfn.startswith("srm://"):
            return "%s lcg-del -b -l -D srmv2 %s --vo cms %s" % (self.setups, self.timeoutOptions, pfn)
        elif pfn.startswith("file:"):
            return "/bin/rm -f %s" % pfn.replace("file:", "", 1)
        else:
            return StageOutImpl.createRemoveFileCommand(self, pfn)

    def createStageOutCommand(self, sourcePFN, targetPFN, options=None, checksums=None):
        """
        _createStageOutCommand_

        Build an srmcp command

        """
        result = "#!/bin/sh\n"

        # check if we should use the grid UI from CVMFS
        useCVMFS = False
        if options == "cvmfs":
            options = None
            useCVMFS = True

        copyCommand = "lcg-cp -b -D srmv2 --vo cms --srm-timeout 2400 --sendreceive-timeout 2400 --connect-timeout 300 --verbose"
        if options != None:
            copyCommand += " %s " % options
        copyCommand += " %s " % sourcePFN
        copyCommand += " %s 2> stageout.log" % targetPFN

        # for CVMFS skip the usual environment setup and use grid.cern.ch instead
        # also open a sub-shell and remove the cms.cern.ch path elements
        if useCVMFS:
            result += "(\n"

            result += "echo Modifying PATH and LD_LIBRARY_PATH to remove /cvmfs/cms.cern.ch elements\n"
            result += "export PATH=`echo $PATH | sed -e 's+:*/cvmfs/cms.cern.ch/[^:]*++'g`\n"
            result += "export LD_LIBRARY_PATH=`echo $LD_LIBRARY_PATH | sed -e 's+:*/cvmfs/cms.cern.ch/[^:]*++'g`\n"

            result += "echo Sourcing CVMFS UI setup script\n"
            # this version dates back to Aug 2013. So, let's get the symlink to the latest one instead
            #result += ". /cvmfs/grid.cern.ch/emi-ui-2.9.0-1_sl5v1/etc/profile.d/setup-cvmfs-ui.sh\n"
            result += ". /cvmfs/grid.cern.ch/emi3ui-latest/etc/profile.d/setup-ui-example.sh\n"

            result += copyCommand
        else:
            result += self.setups
            result += copyCommand

        if _CheckExitCodeOption:
            result += """
            EXIT_STATUS=$?
            cat stageout.log
            echo -e "\nlcg-cp exit status: $EXIT_STATUS"
            if [[ $EXIT_STATUS != 0 ]]; then
                echo "ERROR: lcg-cp exited with $EXIT_STATUS"
                echo "Cleaning up failed file:"
                %s
                exit $EXIT_STATUS
            fi

            """ % self.createRemoveFileCommand(targetPFN)

        if self.stageIn:
            remotePFN, localPFN = sourcePFN, targetPFN.replace("file:", "", 1)
        else:
            remotePFN, localPFN = targetPFN, sourcePFN.replace("file:", "", 1)

        result += "FILE_SIZE=`stat -c %s"
        result += " %s`\n" % localPFN
        result += "echo \"Local File Size is:  $FILE_SIZE\"\n"

        removeCommand = self.createRemoveFileCommand(targetPFN)

        useChecksum = (checksums != None and 'adler32' in checksums and not self.stageIn)

        if useChecksum:
            localAdler32 = str(checksums['adler32']).zfill(8)
            checksumCommand = \
                """
                if [[ "X$SRM_CHECKSUM" != "X" ]]; then
                    if [[ "$SRM_CHECKSUM" == "%s" ]]; then
                        exit 0
                    else
                        echo "ERROR: Checksum Mismatch between local and SE"
                        echo "Cleaning up failed file"
                        %s
                        exit 60311
                    fi
                fi
                exit 0
                """ % (localAdler32, removeCommand)
        else:
            checksumCommand = "exit 0"

        metadataCheck = \
            """
            LCG_OUTPUT=`lcg-ls -l -b -D srmv2 %s %s 2>/dev/null`
            SRM_SIZE=`echo "$LCG_OUTPUT" | awk 'NR==1{print $5}'`
            SRM_CHECKSUM=`echo "$LCG_OUTPUT" | sed -nr 's/^.*\s([a-f0-9]{8})\s*\([aA][dD][lL][eE][rR]32\)\s*$/\\1/p'`
            echo "Remote File Size is: $SRM_SIZE"
            echo "Remote Checksum is:  $SRM_CHECKSUM"
            if [[ $SRM_SIZE == $FILE_SIZE ]]; then
                %s
            else
                echo $LCG_OUTPUT
                echo "ERROR: Size Mismatch between local and SE. Cleaning up failed file..."
                %s
                exit 60311
            fi
            """ % (self.timeoutOptions, remotePFN, checksumCommand, removeCommand)
        result += metadataCheck

        # close sub-shell for CVMFS use case
        if useCVMFS:
            result += ")\n"

        return result

    def createDebuggingCommand(self, sourcePFN, targetPFN, options=None, checksums=None):
        """
        Debug a failed lcg-via smrv copy command for stageOut, without re-running it,
        providing information on the environment and the certifications

        :sourcePFN: str, PFN of the source file
        :targetPFN: str, destination PFN
        :options: str, additional options for copy command
        :checksums: dict, collect checksums according to the algorithms saved as keys
        """
        # Build the command for debugging purposes
        copyCommand = "lcg-cp -b -D srmv2 --vo cms --srm-timeout 2400 --sendreceive-timeout 2400 --connect-timeout 300 --verbose"
        if options != None:
            copyCommand += " %s " % options
        copyCommand += " %s " % sourcePFN
        copyCommand += " %s 2> stageout.log" % targetPFN

        result = self.debuggingTemplate.format(copy_command=copyCommand, source=sourcePFN, destination=targetPFN)
        return result

    def removeFile(self, pfnToRemove):
        """
        _removeFile_

        CleanUp pfn provided

        """
        command = "%s lcg-del -b -l -D srmv2 %s --vo cms %s" % (self.setups, self.timeoutOptions, pfnToRemove)
        self.executeCommand(command)


registerStageOutImpl("srmv2-lcg", LCGImpl)



"""
provide the map between configuration of web tool and
DBFactory parameters
"""
from WMCore.Database.ConfigDBMap import ConfigDBMapInterface

class ConfigDBMap(ConfigDBMapInterface):

    def __init__(self, config):
        assert hasattr(config, 'CoreDatabase'), "No database configured"
        assert hasattr(config.CoreDatabase, 'connectUrl'), "No database url configured"
        self.dbUrl = config.CoreDatabase.connectUrl
        self.option = {}
        if hasattr(config.CoreDatabase, "socket"):
            self.option["unix_socket"] = config.CoreDatabase.socket
        if hasattr(config.CoreDatabase, "engineParameters"):
            self.option['engine_parameters'] = config.CoreDatabase.engineParameters

    def getDBUrl(self):
        return self.dbUrl

    def getOption(self):
        return self.option



# -*- coding: utf-8 -*-
"""
_GetJobSlots_

MySQL implementation of Locations.GetJobSlots
Created on Mon Jun 18 15:26:58 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetJobSlots(DBFormatter):
    sql = """SELECT pending_slots, running_slots FROM  wmbs_location
             WHERE site_name = :location
             """

    def execute(self, siteName, conn = None, transaction = False):
        binds = {"location": siteName}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        formatted = self.formatDict(result)

        return formatted



#!/usr/bin/env python
"""
_GetCompletedByFileList_

Oracle implementation of Subscription.IsFileCompleted
"""

from WMCore.WMBS.MySQL.Subscriptions.GetCompletedByFileList import \
     GetCompletedByFileList as GetCompletedByFileListMySQL

class GetCompletedByFileList(GetCompletedByFileListMySQL):
    pass
